#!/bin/bash

cfg_build_configs() {
  declare -a build_configs=($1)

  log_info "Generating default configuration files ${build_configs[*]}"
  if ngcpcfg --no-validate --no-action-failure --no-check-origin build "${build_configs[@]}" ; then
    log_info "OK generating default configuration"
  else
    log_die "Error running 'ngcpcfg --no-check-origin build'"
    exit 1
  fi
}


cfg_schema_initialise() {
  CONFIG_DIR="$(mktemp -d)"
  export CONFIG_DIR

  if ${PRO} || ${CARRIER} ; then
    revname='0001_init_config_pro.up'
  else
    revname='0001_init_config_ce.up'
  fi
  log_info "Initialising config.yml using ${revname}"
  if ! /usr/share/ngcp-cfg-schema/cfg_scripts/init/${revname} /dev/null "${CONFIG_DIR}/config.yml" ; then
    log_die "could not initialise '${CONFIG_DIR}/config.yml'"
    exit 1
  fi
  log_info "Securing config.yml to allow read for root user only"
  chmod 0600 "${CONFIG_DIR}/config.yml"

  if ${PRO} || ${CARRIER} ; then
    revname='0002_init_constants_pro.up'
  else
    revname='0002_init_constants_ce.up'
  fi
  log_info "Initialising constants.yml using ${revname}"
  if ! /usr/share/ngcp-cfg-schema/cfg_scripts/init/${revname} /dev/null "${CONFIG_DIR}/constants.yml" ; then
    log_die "could not initialise '${CONFIG_DIR}/constants.yml'"
    exit 1
  fi
  log_info "Securing constants.yml to allow read for root user only"
  chmod 0600 "${CONFIG_DIR}/constants.yml"
}

# ugly but necessary as ngcp.cfg_schema isn't available yet
# while bootstrapping the cfg_scripts/init/* scripts
cfg_schema_record_in_db() {
  if ${PRO} || ${CARRIER} ; then
    revnames='0001_init_config_pro.up 0002_init_constants_pro.up'
  else
    revnames='0001_init_config_ce.up 0002_init_constants_ce.up'
  fi

  log_info "Initialising cfg_schema.db"

  # create dbfile
  if ! ngcp-update-cfg-schema -i -f -d "${CONFIG_DIR}" ; then
    log_die 'failed to create cfg_schema.db'
    exit 1
  fi

  for revname in $revnames ; do
    if [[ "$revname" =~ ^([0-9]+)_ ]] ; then
      rev=$(echo "${BASH_REMATCH[1]}" | sed -r 's/^0*//')
    fi

    log_info "Recording ${revname}:${rev} in cfg_schema:"
    if sqlite3 "${CONFIG_DIR}/cfg_schema.db" "insert into cfg_schema(revision, node) values ('${rev}', \"${HNAME}\");" ; then
      log_info "OK recording ${revname}:${rev} in cfg_schema"
    else
      log_die "failed to record revision script in ${CONFIG_DIR}/cfg_schema.db"
      exit 1
    fi
  done
}

cfg_apply_cfg_schema() {
  log_info "Executing ngcp-update-cfg-schema"
  ngcp-update-cfg-schema -f -d "${CONFIG_DIR}"
}

# exclude files from etckeeper
cfg_exclude_etc_files() {
  for ign in "${ETCKEEPER_IGNORE[@]}" ; do
    if [ -e "$ign" ] ; then
      log_info "Removed $ign from etckeeper repo"
      (
        set -e
        cd /etc
        git rm --cached "$ign" || true
      )
    fi
    if ! grep -q "${ign##/etc/}" /etc/.gitignore ; then
      echo "${ign##/etc/}" >> /etc/.gitignore
      log_info "Added '${ign##/etc/}' to /etc/.gitignore"
    fi
  done
  (
    set -e
    cd /etc
    git add .gitignore
    if cfg_is_git_clean ".gitignore"; then
      log_info "Git repo /etc is clean"
    else
      log_info "Committing .gitignore file"
      git commit -m 'add ETCKEEPER_IGNORE files to .gitignore' >/dev/null
    fi
  )
}

cfg_ngcpcfg_initialise() {
  log_info "Initializing the configuration framework..."
  if ngcpcfg initialise --without_shared ; then
    log_info "OK initializing the configuration framework"
  else
    log_die "Error running 'ngcpcfg initialise'"
    exit 1
  fi
}

cfg_amazon_tweaks() {
  if ! check_running_on_aws ; then
    log_info "This doesn't look like an EC2 instance, skipping system_adjust_aws_settings"
    return 0
  fi

  log_info "This looks like an EC2 instance, applying AWS specific changes."

  declare -a pkgs_installed
  pkgs_installed+=(augeas-tools)
  pkgs_installed+=(facter)
  pkgs_installed+=(cloud-initramfs-growroot)
  pkgs_installed+=(cloud-init)
  pkgs_installed+=(ngcp-linux-cloud-amd64)
  log_info "Making sure we have essential packages installed:"
  if env "${APT_ENV[@]}" apt-get "${APT_PARAMS[@]}" install "${pkgs_installed[@]}" ; then
    log_info "OK"
  else
    log_die "Error installing packages."
    exit 1
  fi

  declare -a pkgs_purge
  pkgs_purge+=(unattended-upgrades)
  log_info "Making sure we do not have certain packages installed:"
  if env "${APT_ENV[@]}" apt-get "${APT_PARAMS[@]}" remove --purge "${pkgs_purge[@]}" ; then
    log_info "OK"
  else
    log_die "Error purging packages."
    exit 1
  fi

  log_info "Adding cloud-init services to ngcpcfg's systemd.custom_preset"
  declare -a cloud_services
  cloud_services+=(cloud-config.service)
  cloud_services+=(cloud-final.service)
  cloud_services+=(cloud-init-local.service)
  cloud_services+=(cloud-init.service)
  for cloud_service in "${cloud_services[@]}" ; do
    ngcpcfg set "${CONFIG_YML}" "systemd.custom_preset.APPEND='${cloud_service}'"
  done
  log_info "OK"

  log_info "Enabling sshd to listen on all IPv4 + IPv6 ports via ngcpcfg's sshd.listen_addresses"
  ngcpcfg set "${CONFIG_YML}" 'sshd.listen_addresses=["0.0.0.0", "::"]'
  log_info "OK"

  if [ -n "$(augtool 'match /files/etc/fstab/*[file="/"]/opt noatime')" ] ; then
    log_info "Mount option 'noatime' already in usage for rootfs, nothing to do."
  else
    log_info "Enabling 'noatime' mount option for rootfs: "
    cat << EOT | augtool >/dev/null
insert opt after /files/etc/fstab/*[file="/"]/opt[last()]
set /files/etc/fstab/*[file="/"]/opt[last()] noatime
save
EOT
    log_info "OK"
  fi

  log_info "Checking for public IPv4 address using facter."
  local PUBLIC_IP
  PUBLIC_IP=$(facter ec2_metadata.public-ipv4 2>/dev/null)

  log_info "Checking for primary network device using facter."
  local PRIMARY_NETDEVICE
  PRIMARY_NETDEVICE=$(facter networking.primary 2>/dev/null)

  if [ -z "${PRIMARY_NETDEVICE}" ] ; then
    echo "Couldn't identify primary network device with facter, falling back to ip route information"
    PRIMARY_NETDEVICE=$(ip route | grep '^default' | sed 's/.* dev //')
  fi

  if ! is_valid_ipv4 "${PUBLIC_IP}" ; then
    log_warn "Could not retrieve IPv4 address, not running ngcp-network to adjust advertised IP address."
  else
    log_info "Setting ${PUBLIC_IP} as advertised IP address on ${PRIMARY_NETDEVICE}."
    ngcp-network --set-interface="${PRIMARY_NETDEVICE}" --advertised-ip="${PUBLIC_IP}"
  fi

  if [ -f /etc/network/cloud-ifupdown-helper ] ; then
    log_info "Detected cloud-ifupdown-helper, using dynamic network configuration"
    log_info "Generating /etc/ngcp-config/templates/etc/network/interfaces.customtt.tt2"
    tee > /etc/ngcp-config/templates/etc/network/interfaces.customtt.tt2 << EOF
# Generated by ngcpcfg via customtt template for usage on AWS
# Based on https://salsa.debian.org/cloud-team/debian-cloud-images/blob/master/config_space/files/etc/network/interfaces/CLOUD

# Include files from /etc/network/interfaces.d:
source-directory /etc/network/interfaces.d

# Cloud images dynamically generate config fragments for newly
# attached interfaces. See /etc/udev/rules.d/75-cloud-ifupdown.rules
# and /etc/network/cloud-ifupdown-helper. Dynamically generated
# configuration fragments are stored in /run:
source-directory /run/network/interfaces.d

EOF
  else
    log_info "No cloud-ifupdown-helper present, using static network configuration"
    log_info "Generating /etc/ngcp-config/templates/etc/network/interfaces.customtt.tt2"
    tee > /etc/ngcp-config/templates/etc/network/interfaces.customtt.tt2 << EOF
# Generated by ngcpcfg via customtt template for usage on AWS
# Based on https://sources.debian.org/src/debian-cloud-images/latest/config_space/scripts/EC2/10-network

auto lo
iface lo inet loopback
iface lo inet6 loopback

iface eth0 inet dhcp
allow-hotplug eth0
iface eth0 inet6 manual
  up /usr/local/sbin/inet6-ifup-helper
  down /usr/local/sbin/inet6-ifup-helper
iface eth1 inet dhcp
allow-hotplug eth1
iface eth1 inet6 manual
  up /usr/local/sbin/inet6-ifup-helper
  down /usr/local/sbin/inet6-ifup-helper
iface eth2 inet dhcp
allow-hotplug eth2
iface eth2 inet6 manual
  up /usr/local/sbin/inet6-ifup-helper
  down /usr/local/sbin/inet6-ifup-helper
iface eth3 inet dhcp
allow-hotplug eth3
iface eth3 inet6 manual
  up /usr/local/sbin/inet6-ifup-helper
  down /usr/local/sbin/inet6-ifup-helper
iface eth4 inet dhcp
allow-hotplug eth4
iface eth4 inet6 manual
  up /usr/local/sbin/inet6-ifup-helper
  down /usr/local/sbin/inet6-ifup-helper
iface eth5 inet dhcp
allow-hotplug eth5
iface eth5 inet6 manual
  up /usr/local/sbin/inet6-ifup-helper
  down /usr/local/sbin/inet6-ifup-helper
iface eth6 inet dhcp
allow-hotplug eth6
iface eth6 inet6 manual
  up /usr/local/sbin/inet6-ifup-helper
  down /usr/local/sbin/inet6-ifup-helper
iface eth7 inet dhcp
allow-hotplug eth7
iface eth7 inet6 manual
  up /usr/local/sbin/inet6-ifup-helper
  down /usr/local/sbin/inet6-ifup-helper
iface eth8 inet dhcp
allow-hotplug eth8
iface eth8 inet6 manual
  up /usr/local/sbin/inet6-ifup-helper
  down /usr/local/sbin/inet6-ifup-helper
EOF
  fi
}

cfg_build_templates() {
  cfg_build_configs

  log_info "Saving config files..."
  cfg_exclude_etc_files
  pushd /etc &>/dev/null
  if cfg_is_git_clean; then
    log_info "/etc repo is clean"
  else
    etckeeper commit "Initial config"
  fi
  popd &>/dev/null
  ngcpcfg --no-db-sync commit "Config files were created"
  log_info "Done saving config files"
}

cfg_save_factory_defaults() {
  if [ -f "${CONFIG_YML}" ] ; then
    cp "${CONFIG_YML}" "${CONFIG_YML}.factory_default"
    chown root:root "${CONFIG_YML}.factory_default"
    chmod 0400 "${CONFIG_YML}.factory_default"
  else
    log_die "Main config file does not exist"
    exit 1
  fi

  if [ -f "${NETWORK_YML}" ] ; then
    cp "${NETWORK_YML}" "${NETWORK_YML}.factory_default"
    chown root:root "${NETWORK_YML}.factory_default"
    chmod 0400 "${NETWORK_YML}.factory_default"
  else
    log_die "Main network config file does not exist"
    exit 1
  fi
}

cfg_generate_certs() {
  log_info "Creating group ssl-cert and add users into it"
  mkdir -p "$CERT_PATH"
  find "$CERT_PATH" -type d -exec chmod 0750 {} \;
  chown root:ssl-cert "$CERT_PATH"
  for u in $CERT_USERS ; do
    if ! groups "$u"| grep -q ssl-cert 2>/dev/null ; then
      log_info "Adding user $u to ssl-cert"
      adduser "$u" ssl-cert
    fi
  done

  log_info "Generating openssl keys"
  if [ ! -x /usr/share/ngcp-panel-tools/generate_ssl_keys.sh ]; then
    log_error "No /usr/share/ngcp-panel-tools/generate_ssl_keys.sh found"
    return 1
  fi

  if [ -f "${CERT_PATH}/myserver.crt" ] && \
     [ -f "${CERT_PATH}/myserver.csr" ] && \
     [ -f "${CERT_PATH}/myserver.key" ]
  then
    log_info "Self-signed WEB keys 'myserver' already exist, continue..."
  else
    log_info "Generating self-signed WEB keys 'myserver':"
    if /usr/share/ngcp-panel-tools/generate_ssl_keys.sh "${CERT_PATH}" /usr/share/ngcp-panel-tools myserver ; then
      log_info "OK generating self-signed 'myserver' keys"
    else
      log_error "Problem generating self-signed 'myserver' keys"
      return 1
    fi
  fi

  if [ -f "${CERT_PATH}/api_ca.crt" ] && \
     [ -f "${CERT_PATH}/api_ca.key" ]
  then
    log_info "REST auth keys 'api_ca' already exist, continue..."
  else
    log_info "Generating REST auth keys 'api_ca':"
    if SKIP_CSR=true /usr/share/ngcp-panel-tools/generate_ssl_keys.sh "${CERT_PATH}" ; then
      chmod a+r "${CERT_PATH}"/api_ca.*
      log_info "OK generating REST auth keys 'api_ca'"
    else
      log_error "Problem generating REST auth keys 'api_ca'"
      return 1
    fi
  fi

  if [ -f "${CERT_PATH}/dhparam.pem" ] ; then
    log_info "dhparams.pem already exists, continue..."
  else
    log_info "Generating dhparams.pem (this is going to take a long time)"
    openssl dhparam -dsaparam -out "${CERT_PATH}/dhparam.pem" 4096
    chown root:ssl-cert "${CERT_PATH}/dhparam.pem"
    chmod 640 "${CERT_PATH}/dhparam.pem"
    log_info "Done generating dhparams.pem"
  fi

  cfg_save_defaults "ssl certs" "--no-db-sync"
}

cfg_save_defaults() {
  local message=${1:-}
  declare -a opts=(${2:-})

  if [ "$(ngcpcfg diff 2>/dev/null | wc -l)" != "0" ] ; then
    log_info "Committing config changes for $message"
    ngcpcfg "${opts[@]}" commit "changes for $message"
    etckeeper commit "changes for $message" || true
    log_info "Done"
  fi
}

cfg_create_influx_databases() {
  log_info "Executing ngcp-influxdb-create to generate InfluxDB databases"

  # Make sure the InfluxDB daemon is running so that we can write to it.
  # We can not use restart here as it always returns 1
  service influxdb stop
  service influxdb start

  # InfluxDB can be slow on a start, we need to wait for it
  local TRY=60
  while ! influx --execute "show databases" >/dev/null 2>&1 ; do
    if [ $TRY -gt 0 ] ; then
      TRY=$((TRY-1))
      log_info "Waiting for influxdb to start ($TRY seconds)..."
      sleep 1
    else
      log_die "influxdb failed to start"
      exit 1
    fi
  done

  ngcp-influxdb-create
}

cfg_systemctl_preset_all() {
  if "${DOCKER_BUILD}"; then
    log_info "It is Docker build, no need to configure preset"
    return 0
  fi
  log_info "Removing /etc/systemd/system-preset/01-installer.preset"
  rm -f /etc/systemd/system-preset/01-installer.preset
  log_info "Calling systemd preset to configure services on the next boot:"
  systemctl preset-all
}

cfg_set_network_config_ce() {
  log_info "Update ${NETWORK_YML} with real CE network parameters"

  if [[ "${USE_EXISTING_YML}" =~ network ]]; then
    log_info "It is specified to use prepared network.yml, do not change in ${FUNCNAME[0]}"
    return
  fi

  ngcp-network --set-interface="${EXTERNAL_DEV}" \
    --ip="${EADDR}" --netmask="${EXTERNAL_NETMASK}" --gateway="${GW}" --hwaddr=auto

  for entry in ${NAMESERVER} ; do
    ngcp-network --set-interface="${EXTERNAL_DEV}" --dns="${entry}"
  done

  vm_net_configuration 'spce'
}

cfg_build_net_configs() {
  log_info "Generating default configuration files /etc/network/interfaces /etc/hosts /etc/iproute2/rt_tables"
  cfg_build_configs "/etc/network/interfaces /etc/hosts /etc/iproute2/rt_tables"
}

cfg_apply_configs_docker() {
  log_info "It is Docker build, can't apply configuration due to systemctl limitation"

  log_info "Building configuration"
  if ! ngcpcfg build ; then
    log_die "Can not build configuration"
    exit 1
  fi

  log_info "Committing configuration"
  if ! ngcpcfg commit "Installer building/committing configuration" ; then
    log_die "Can not commit configuration"
    exit 1
  fi
}

cfg_apply_configs_default() {
  log_info "Applying configuration"
  if ! ngcpcfg apply --force-all-services "Installer applying configuration" ; then
    log_die "Can not apply configuration"
    exit 1
  fi
}

cfg_apply_configs() {
  if "${DOCKER_BUILD}"; then
    cfg_apply_configs_docker
  else
    cfg_apply_configs_default
  fi
}

get_underlying_net_dev() {
  local device
  device="${1}"
  if [[ -z "${device}" ]]; then
    log_die "Interface name has to be passed to get_underlying_net_dev"
    exit 1
  fi

  local underlying_device
  underlying_device="$( ip a show "${device}" | sed -r 's/.+vlan[0-9]+@(.+):.+$/\1/;t;d' )"
  if [[ -z "${underlying_device}" ]]; then
    log_die "The underlying device was not detected"
    exit 1
  fi
  echo "${underlying_device}"

  return 0
}

cfg_normalize_net_dev() {
  if [[ "${EXTERNAL_DEV}" =~ ^vlan ]]; then
    EXTERNAL_DEV="$( get_underlying_net_dev "${EXTERNAL_DEV}" )"
  fi
}

cfg_common_configuration() {
  if [[ "${USE_EXISTING_YML}" =~ config ]]; then
    log_info "It is specified to use prepared config.yml, do not change in ${FUNCNAME[0]}"
    return
  fi

  log_info "Save size of 'code' partitions to ${CONFIG_YML}"
  ngcpcfg set "${CONFIG_YML}" "bootenv.netscript.fallbackfssize=${FALLBACKFS_SIZE}"
  ngcpcfg set "${CONFIG_YML}" "bootenv.netscript.rootfssize=${ROOTFS_SIZE}"

  log_info "Save size of 'swapfile' to ${CONFIG_YML}"
  ngcpcfg set "${CONFIG_YML}" "bootenv.netscript.swapfilesize=${SWAPFILE_SIZE_MB}"

  local POOL_SIZE_MB
  if ${PRO} || ${CARRIER}; then
    log_info "Preparing dynamic configuration settings for MySQL:"
    local TOTAL_MEM
    TOTAL_MEM=$(awk '/^MemTotal:/ {print $2}' /proc/meminfo)

    log_info "    Total memory detected: ${TOTAL_MEM}"
    [ -z "${INNODB_BUFFER_POOL_FACTOR}" ] && INNODB_BUFFER_POOL_FACTOR="0.5"

    log_info "    Innodb poolsize factor set to: ${INNODB_BUFFER_POOL_FACTOR}"
    POOL_SIZE_MB=$(echo | awk -v tm="${TOTAL_MEM}" -v mf=${INNODB_BUFFER_POOL_FACTOR} '{print int(tm*mf/1000)}')
    POOL_SIZE_MB="${POOL_SIZE_MB}M"
  else
    # in CE systems we hardcode the innodb_pool_size setting
    log_info "CE system detected, setting static configuration for MySQL innodb_pool_size:"
    POOL_SIZE_MB="64M"
  fi
  log_info "Save innodb_pool_size ${POOL_SIZE_MB} to ${CONFIG_YML}"
  ngcpcfg set "${CONFIG_YML}" "database.bufferpoolsize=${POOL_SIZE_MB}"

  # we need to decrease default high performance config
  # for low performance virtual environment (e.g. Vagrant)
  if "${ADJUST_FOR_LOW_PERFORMANCE}" ; then
    log_info "Enabling low performance mode"
    /usr/sbin/ngcp-toggle-performance-config --low-performance --skip-commit
  fi
  if "${ENABLE_VM_SERVICES}" ; then
    # we do enable some services on Vagrant VMs for testing
    enable_vm_services
  fi

  if [[ -n "${NGCP_PPA}" ]]; then
    local old_IFS="${IFS}"
    IFS=','
    declare -a ppas=( ${NGCP_PPA} )
    for ppa in "${ppas[@]}"; do
      ngcpcfg set "${CONFIG_YML}" bootenv.ppa.APPEND="{'name','${ppa}'}"
    done
    IFS="${old_IFS}"
  fi
}

cfg_is_git_clean() {
  log_debug "call 'git status --porcelain=v2 $*' in $(pwd). it must have no output"
  if [[ -z "$(git status --porcelain=v2 "$@")" ]]; then
    return 0
  else
    return 1
  fi
}

cfg_place_yml_files() {
  log_info "Copying '${CONFIG_DIR}/cfg_schema.db' to '${CFG_SCHEMA_FILE}'"
  cp "${CONFIG_DIR}/cfg_schema.db" "${CFG_SCHEMA_FILE}"

  if [[ "${USE_EXISTING_YML}" =~ config ]]; then
    log_info "USE_EXISTING_YML config is specified so config.yml is not copied"
  else
    log_info "Copying '${CONFIG_DIR}/config.yml' to '${CONFIG_YML}'"
    cp "${CONFIG_DIR}/config.yml" "${CONFIG_YML}"
  fi

  if [[ "${USE_EXISTING_YML}" =~ constants ]]; then
    log_info "USE_EXISTING_YML constants is specified so constants.yml is not copied"
  else
    log_info "Copying '${CONFIG_DIR}/constants.yml' to '${CONSTANTS_YML}'"
    cp "${CONFIG_DIR}/constants.yml" "${CONSTANTS_YML}"
  fi

  if [[ "${USE_EXISTING_YML}" =~ network ]]; then
    log_info "USE_EXISTING_YML network is specified so network.yml is not copied"
  else
    log_info "Copying '${CONFIG_DIR}/network.yml' to '${NETWORK_YML}'"
    cp "${CONFIG_DIR}/network.yml" "${NETWORK_YML}"
  fi
}

cfg_ask_user() {
  local message="$*"
  log_info "${message} [y/N]"
  local ACCEPT
  if [[ "${FORCE}" == 'yes' ]]; then
    log_info "FORCE mode is used, accepting"
    ACCEPT='y'
  else
    read -r ACCEPT
  fi
  if [[ "${ACCEPT,,}" != 'y' ]] ; then
    log_info "Exiting as requested"
    exit 1
  fi
}

cfg_generate_ngcp_panel_keys() {
  log_info "Creating asymmetric keys for ngcp-panel"
  ngcp-panel-create-keys
  cfg_save_defaults "ngcp-panel keys" "--no-db-sync"
}
