#!/bin/bash

mysql_apply_grants() {
  xtracestate="$(shopt -po xtrace || true)"
  echo "Disabling bash xtrace behavior to not leak passwords."
  set +x

  local sipwise_pass
  sipwise_pass="$(ngcpcfg get 'credentials.mysql.system.p')"
  if ! mysql -e "GRANT ALL PRIVILEGES ON *.* TO sipwise@localhost IDENTIFIED BY '${sipwise_pass}' WITH GRANT OPTION;" ; then
    log_die "grants for user sipwise on localhost could not be applied."
    exit 1
  fi

  eval "${xtracestate}"
}

mysql_apply_db_schema() {
  AUTOMATED_INSTALL_MODE=1 SKIP_SYNC_DB=1 /usr/sbin/ngcp-update-db-schema -f
}

mysql_replication() {
  id=1
  REPLICATION=false

  if ! $DPL_MYSQL_REPLICATION ; then
    return
  fi

  # wait-check for replication to finish, init script is unreliable :(
  while [ "$id" -lt 25 ] ; do

    local free_memory
    free_memory=$(awk '/^MemFree:/ {print $2}' /proc/meminfo)
    if [ "$free_memory" -lt 100000 ] ; then
      log_warn "looks like system is running out of memory (only $free_memory free)."
      log_warn "MySQL errors like 'Table ... doesn't exist' showing up? Try increasing RAM."
    fi

    local rc=0
    ngcp-mariadb-replication-check || rc=$?
    if [ $rc -eq 3 ] ; then
      log_info "Replication is catching up, waiting..."
    else
      log_info "Testing MySQL replication status (test ${id}/24)..."
      sleep 5 # be as conservative as possible
      id=$((id + 1))

      if mysql --defaults-extra-file="${SIPWISE_EXTRA_CNF}" -e 'use accounting' ; then
        mysql --defaults-extra-file="${SIPWISE_EXTRA_CNF}" -e "show slave status \G;"
        log_info "OK"
        REPLICATION=true
        break
      fi
    fi

    sleep 5
  done

  if ! $REPLICATION ; then
    log_die "Error setting up mysql replication."
    exit 1
  fi
}

mysql_restart_sync() {
  if ! "${JOIN_CLUSTER}" ; then return ; fi

  # Check if there is alive peer. If no - it means that the 1st node from the pair is set up.
  if ! mysql --defaults-extra-file="${SIPWISE_EXTRA_CNF}" -h "${PEER}" -e "exit" ; then
    return
  fi

  if ! "${DPL_MYSQL_REPLICATION}" ; then
    return
  fi
  systemctl start mysql || true

  /usr/share/ngcp-ngcpcfg/helper/check-for-mysql

  log_info "Don't initialize the databases"

  /usr/sbin/ngcp-sync-db --force --verbose --ssh-tunnel=33125 --local-user=root --local-host=localhost --fix-db-schema --set-local-grants

  # renable normal sync to the DB
  unset NO_DB_SYNC

  mysql_replication

  mysql_apply_db_schema
}

mysql_start_service() {
  service mysql start || true
  /usr/share/ngcp-ngcpcfg/helper/check-for-mysql
}

mysql_init_db_apply() {
  mysql_start_service
  mysql_apply_grants
  mysql_apply_db_schema
}

mysql_init_db() {
  mysql_initialize_db
  if "${CE}" ; then
    mysql_init_db_apply
  elif "${JOIN_CLUSTER}" ; then
    # Check if there is alive peer. If yes - it means that the 2nd node from the pair is set up.
    if mysql --defaults-extra-file="${SIPWISE_EXTRA_CNF}" -h "${PEER}" -e "exit" ; then
      # we do not have the NGCP db yet, no syncs to the DB for now
      export NO_DB_SYNC=1
    else
      mysql_init_db_apply
    fi
  else
    mysql_init_db_apply
  fi

  if ${CARRIER} && [ "$CROLE" = "proxy" ] ; then
    # initialise the R/O mysql db
    if [ ! -d "/ngcp-data/mysql2" ] ; then
      declare -a mysql2_install_options
      mysql2_install_options+=("--defaults-file=/etc/mysql/conf.d/my2.cnf")
      mysql2_install_options+=("--innodb_encrypt_tables=off")
      mysql2_install_options+=("--innodb_encrypt_log=0")
      mysql2_install_options+=("--encrypt_binlog=0")
      log_info "Creating mysql2 instance in /ngcp-data/mysql2"
      mysql_install_db "${mysql2_install_options[@]}"
      log_info "Done"
    fi
  fi
}

mysql_sync_central_db() {
  if [ "$CROLE" = "proxy" ] ; then
    # start the R/O mysql node
    systemctl restart mariadb@2.service || true
    /usr/share/ngcp-ngcpcfg/helper/check-for-mysql 60 /etc/mysql/conf.d/my2.cnf
    NGCP_TYPE="carrier" NGCP_IS_PROXY="yes" AUTOMATED_INSTALL_MODE=1 /usr/sbin/ngcp-update-db-schema -f
  fi
}

mysql_mariadb_migration_configs() {
  if [ ! -r /etc/mysql/my.cnf ]; then
    log_die "ERROR: cannot read /etc/mysql/my.cnf, cannot continue!"
    exit 1
  fi

  if [ -h /etc/mysql/my.cnf ]; then
    log_info "/etc/mysql/my.cnf is symbolic link, converting to the plain config"
    tmpfile=$(mktemp)
    cat /etc/mysql/my.cnf > "${tmpfile}"
    rm /etc/mysql/my.cnf
    mv "${tmpfile}" /etc/mysql/my.cnf
    chmod 0444 /etc/mysql/my.cnf
    chown root:root /etc/mysql/my.cnf
    ln -sf /etc/mysql/my.cnf /etc/alternatives/my.cnf
    log_info "Done"
  else
    log_info "/etc/mysql/my.cnf is NOT symbolic link, all OK here"
  fi

  FILES=(/etc/mysql/my.cnf.migrated /etc/mysql/my.cnf.fallback /etc/mysql/my.cnf.dpkg-dist)

  for file in "${FILES[@]}" ; do
    if [ -f "${file}" ]; then
      log_info "Removing unnecessary '${file}'"
      rm "${file}"
    else
      log_info "File '${file}' is missing, OK here"
    fi
  done

  if [ -d /etc/mysql/mariadb.conf.d ]; then
    log_info "Removing not-yet supported /etc/mysql/mariadb.conf.d/"
    rm -rf /etc/mysql/mariadb.conf.d
  fi
}

mysql_sync_grants_constants() {
  if "${JOIN_CLUSTER}" ; then return ; fi

  echo "Syncing grants/constants"
  ngcp-sync-grants
  ngcp-sync-constants
  echo "Done"
}

mysql_initialize_db() {
  local mysql_dir='/ngcp-data/mysql'
  local mysql_default_dir='/var/lib/mysql'

  service mysql stop
  if [[ -d "${mysql_dir}" ]]; then
    log_warn "Directory ${mysql_dir} already exists."
    cfg_ask_user "Are you sure you want to delete it and reinitialize mysql db instance?"
    rm -rf "${mysql_dir}"
  fi

  mkdir -p "${mysql_dir}"
  chown mysql:mysql "${mysql_dir}"

  log_info "Creating backward compatibility link to new data storage"
  rm -rf "${mysql_default_dir}"
  ln -s "${mysql_dir}" "${mysql_default_dir}"

  declare -a mysql_install_options
  mysql_install_options+=("--defaults-file=/etc/mysql/my.cnf")
  mysql_install_options+=("--verbose")
  mysql_install_options+=("--skip-test-db")
  mysql_install_options+=("--innodb_encrypt_tables=off")
  mysql_install_options+=("--innodb_encrypt_log=0")
  mysql_install_options+=("--encrypt_binlog=0")
  log_info "Invoking mysql_install_db with '--skip-test-db' and '--innodb_encrypt_tables=off --innodb_encrypt_log=0 --encrypt_binlog=0' options (mariadb >=10.3.34)"
  mysql_install_db "${mysql_install_options[@]}"

  mysql_start_service
}
