#!/bin/bash

packages_enable_policy_rc_d() {
  local POLICY_RC_D_FILE="/usr/sbin/policy-rc.d"

  if [ -f "${POLICY_RC_D_FILE}" ]; then
    log_warn "policy_rc.d file '${POLICY_RC_D_FILE}' already available, overwriting it"
  fi

  log_info "Creating file '${POLICY_RC_D_FILE}'"
  cat > "${POLICY_RC_D_FILE}" << "EOF"
#!/bin/sh
# ngcp-installer-marker
if [ "$2" = "stop" ] ; then
  echo "[$(date)] NGCP policy-rc.d ngcp-installer wrapper, allowed action: $@"
  exit 0
else
  echo "[$(date)] NGCP policy-rc.d ngcp-installer wrapper, denied action: $@"
  exit 101
fi
EOF

  chmod 0775 "${POLICY_RC_D_FILE}"
}

packages_fetch_pkglist() {
  declare -a _apt_env=($1)
  declare -a _apt_param=($2)
  local _str="$3"

  log_info "Fetching package list ${_str}"
  # work around MT#16183 -> https://bugs.debian.org/776152
  local tmpfile
  tmpfile=$(mktemp)
  if ! env "${_apt_env[@]}" apt-get "${_apt_param[@]}" update 2>&1 | tee -a "${tmpfile}" ; then
    log_die "Error while retrieving package information ${_str}"
    exit 1
  fi

  if grep -q 'Failed to fetch' "${tmpfile}" ;  then
    log_die "failed to fetch repository files, exiting."
    exit 1
  fi
  # workaround TT#16509
  if grep -Eq '^E: |^Err ' "${tmpfile}" ;  then
    log_die "repository metadata update failed, exiting."
    exit 1
  fi

  rm -f "${tmpfile}"
  log_info "OK"
}

packages_ppa_sourceslist() {
  if [ -z "${NGCP_PPA}" ] ; then
    log_info "Skipping NGCP PPA  generation as variable NGCP_PPA is not set."
    return 0
  fi

  log_info "Installing ngcp-dev-tools"
  packages_fetch_pkglist
  packages_install_dev_tools

  declare -a PPA_URL_OPT
  PPA_URL_OPT+=("--clean")
  PPA_URL_OPT+=("--force")
  PPA_URL_OPT+=("--url ${SIPWISE_URL}")
  PPA_URL_OPT+=("--without-ngcpcfg")
  echo "Using PPA URL from SIPWISE_URL variable: ${SIPWISE_URL}"

  log_info "Adding ngcp-ppa repositories: $NGCP_PPA"
  if ngcp-ppa "${PPA_URL_OPT[@]}" -k "${NGCP_PPA}" ; then
    log_info "OK"
  else
    log_die "Error while installing ngcp-ppa packages"
    exit 1
  fi

  echo "(system installed using NGCP PPA: ${NGCP_PPA})" >> /etc/sipwise_ngcp_version
}

packages_sourceslist() {
  if [ -z "${SIPWISE_REPO_TYPE:-}" ]; then
    log_die "required variable \$SIPWISE_REPO_TYPE is not defined"
    exit 1
  fi

  log_info "Updating package repository... "
  local TRUNK_PREFIX=''
  if [ "${VERSION}" = "trunk" ] ; then
    SIPWISE_REPO="autobuild"
    TRUNK_PREFIX='release-trunk-'
  else
    SIPWISE_REPO="${SIPWISE_REPO_TYPE}/${VERSION}"
  fi
  mkdir -p /etc/apt/sources.list.d
  local sipwise_list_file="/etc/apt/sources.list.d/sipwise.list"
  cat > "${sipwise_list_file}" << EOF
# NGCP_MANAGED_FILE - do not remove this line if it should be automatically handled

# Sipwise repository
deb [arch=amd64] ${SIPWISE_URL}/${SIPWISE_REPO} ${TRUNK_PREFIX}${DEBIAN_FLAVOUR} main
#deb-src ${SIPWISE_URL}/${SIPWISE_REPO} ${TRUNK_PREFIX}${DEBIAN_FLAVOUR} main
EOF

  # add NGCP PPA repositories if necessary
  packages_ppa_sourceslist
}

packages_aptconfig() {
  mkdir -p /etc/apt/preferences.d

  cat > /etc/apt/preferences.d/sipwise << EOF
Explanation: default Sipwise apt preferences created by ngcp-installer
Package: *
Pin: release o=Sipwise
Pin-Priority: 990
EOF

  mkdir -p /etc/apt/apt.conf.d

  cat > /etc/apt/apt.conf.d/73_acquire_retries << EOF
# NGCP_MANAGED_FILE -- ngcp-installer
Acquire::Retries "3";
EOF
}

packages_install_dev_tools() {
  if [ "${VERSION}" != "trunk" ] && [ -z "${NGCP_PPA}" ] ; then
    log_info "Skippng ngcp-dev-tools installation (it is not a trunk and no \$NGCP_PPA defined)"
    return
  fi

  log_info "Installing ngcp-dev-tools package"
  if env "${APT_ENV[@]}" apt-get "${APT_PARAMS[@]}" install ngcp-dev-tools >> "${LOGFILE}" 2>&1 ; then
    log_info "...installed OK"
  else
    log_die "Error while installing ngcp-dev-tools package"
    exit 1
  fi
}

packages_install_pkglist() {
  # fetch package list
  packages_fetch_pkglist

  local apt_cache_archives='/ngcp-data/cache/apt-archives'
  mkdir -p "${apt_cache_archives}"

  declare -a apt_options=( "${APT_PARAMS[@]}" )
  apt_options+=( "-o" "Dir::Cache::Archives=${apt_cache_archives}" )

  # make sure we have our own keyring available for secure download of further packages
  log_info "Installing package ngcp-archive-keyring ..."
  if ! env "${APT_ENV[@]}" apt-get "${apt_options[@]}" install ngcp-archive-keyring ; then
    log_die "Error while installing ngcp-archive-keyring package"
    exit 1
  fi

  # make sure to retrieve package information INCLUDING our keyring
  packages_fetch_pkglist "" "" "(including ngcp-archive-keyring package)"

  package_check_kernel_update

  # install ngcp metapackage with all necessary dependencies
  log_info "Installing package ${NGCP_METAPACKAGE} ..."
  if ! env "${APT_ENV[@]}" apt-get "${apt_options[@]}" install "${NGCP_METAPACKAGE}" ; then
    log_die "Error while installing ${NGCP_METAPACKAGE}"
    exit 1
  fi

  # make sure all packages are up2date (e.g. lsb-base from our backports repos)
  log_info "Making sure all packages are up to date ..."
  if ! env "${APT_ENV[@]}" apt-get "${apt_options[@]}" upgrade ; then
    log_die "Error upgrading packages"
    exit 1
  fi

  packages_install_dev_tools

  # when a new package version depends on a foreign (not yet installed) package
  # then make sure we also upgrade this package (e.g. linux-image-amd64 from backports)
  log_info "Running aptitude's safe-upgrade ..."
  if ! env "${APT_ENV[@]}" aptitude -y \
     -o DPkg::Options::=--force-confask  -o DPkg::Options::=--force-confdef \
     -o DPkg::Options::=--force-confmiss -o DPkg::Options::=--force-confnew \
     safe-upgrade
  then
    log_die "Error upgrading packages"
    exit 1
  fi

  apt-get clean
  rm -rf "${apt_cache_archives}"
}

package_check_kernel_update() {
  if env "${APT_ENV[@]}" apt-get install --dry-run "${NGCP_METAPACKAGE}" | grep -Eq '^Inst linux-image'; then
    log_info "A new version of the Linux kernel will be installed. Please reboot the system when the installation is done."
    cfg_ask_user "Press 'y' to continue, or any other key to stop the installation."
  fi
}
