#!/bin/bash

system_set_type() {
  if [ -z "${NGCP_TYPE:-}" ]; then
    log_die "required variable \$NGCP_TYPE is not defined, aborting"
    exit 1
  fi

  log_info "Generate a stub ngcp-roles file to define NGCP system type"
  echo "NGCP_TYPE=\"$NGCP_TYPE\"" >/etc/default/ngcp-roles
}

system_set_version() {
  INSTDATE=$(date +"%Y-%m-%d %H:%M:%S")
  echo "System installed. NGCP version $VERSION on $INSTDATE" >> /etc/sipwise_ngcp_version
  echo "$VERSION" > /etc/ngcp_version
}

system_create_log_path() {
  log_info "Creating logs storage on 'data' partition"
  mkdir -p /ngcp-data/
  rm -rf /ngcp-data/logs
  mv /var/log/ngcp /ngcp-data/logs
  ln -s -T /ngcp-data/logs /var/log/ngcp

  mkdir /ngcp-data/logs/old
  chgrp -R adm /ngcp-data/logs
}

system_create_ngcp_data_home() {
  local home_base="/ngcp-data/home"
  echo "Creating ${home_base} for home directories"
  mkdir -p "${home_base}"
  chmod 0755 "${home_base}"
  chown root:root "${home_base}"
}

system_create_ngcp_data_backup() {
  local backup_base="/ngcp-data/backup"
  echo "Creating ${backup_base} for backup directories"
  mkdir -p "${backup_base}"
  chmod 0700 "${backup_base}"
  chown root:root "${backup_base}"

  # legacy location
  ln -s "${backup_base}" "/var/backup"
}

system_create_sipwise_user() {
  log_info "Creating user sipwise"
  if ngcp-support-access --enable ; then
    log_info "OK"
  else
    log_die "Failed to create user sipwise"
    exit 1
  fi
}

system_create_cdr_user() {
  log_info "Creating cdrexport user... "
  "$FILEPATH/create_jail" "$CDRUSER"
  log_info "OK"
}

system_create_backup_user() {
  log_info "Creating ngcp-backup user... "
  log_info
  "$FILEPATH/create_jail" "$BACKUPUSER"
  log_info "OK"
}

system_create_data_dirs() {
  log_info "Creating ngcp-data directories... "

  local dir
  declare -a data_names
  data_names+=(influxdb)
  data_names+=(redis)

  for name in "${data_names[@]}"; do
    dir="/ngcp-data/${name}"
    log_info "Creating ${dir}..."
    mkdir -p "${dir}"
    chown "${name}:${name}" "${dir}"
  done

  data_names=()
  data_names+=(voisniff)

  for name in "${data_names[@]}"; do
    dir="/ngcp-data/${name}"
    log_info "Creating ${dir}..."
    mkdir -p "${dir}"
  done

  if [ -d /var/lib/ngcp-voisniff ]; then
    rmdir /var/lib/ngcp-voisniff
  fi
  ln -s -T /ngcp-data/voisniff /var/lib/ngcp-voisniff

  log_info "OK"
}

system_debsums_init() {
  if [ -x "$(which debsums_init)" ] ; then
    debsums_init
  else
    log_error "debsums_init not present."
  fi
}

system_generate_etc_default_locale() {
  if [ -r /etc/default/locale ] ; then
    log_info "File /etc/default/locale exists already, nothing to do."
  else
    log_info "Generating default /etc/default/locale file."
    cat > /etc/default/locale << EOF
# File generated by ngcp-installer
EOF
    chmod 644 /etc/default/locale
  fi
}

system_print_begin() {
  log_info ""
  log_info "Starting NGCP script ${0} at $(date)"
  log_info ""
}

system_print_done_configuration() {
  log_info ""
  log_info ""
  log_info "System was successfully configured, now you have the best VoIP software."
  log_info ""
  log_info ""
  log_debug "[$(date)] Finished execution of the script ${0}"
}

system_print_done_installation() {
  log_info ""
  log_info ""
  if ${CARRIER} ; then
    log_info "Installation finished. Thanks for choosing Sipwise C5 CARRIER."
  elif ${PRO} ; then
    log_info "Installation finished. Thanks for choosing Sipwise C5 PRO."
  else
    log_info "Installation finished. Thanks for choosing Sipwise C5 Community Edition."
  fi
  log_info "Please reboot the server to continue with configuration."
  log_info ""
  log_info ""
  log_debug "[$(date)] Finished execution of the script ${0}"
}

system_sync_time() {
  log_info "Checking local time"
  if systemctl status systemd-timesyncd.service &>/dev/null; then
    if timedatectl show --property='NTPSynchronized' --value | grep -q 'no'; then
      log_error "The local time is not synced with time server"
      log_die "Sync the time before continuing, otherwise it can damage whole installation"
    fi
  elif systemctl status ntp &>/dev/null; then
    log_info "The ntp service is detected, can't check if the local time is synced"
    log_info "Check if the local time is synced, otherwise it can damage whole installation"
    cfg_ask_user "Is local time synced?"
    log_info "Continue as requested"
  else
    log_error "Can't define which time service is used and if local time is synced"
    log_error "Ensure the local time is synced before continuing, otherwise it can damage whole installation"
    cfg_ask_user "Is local time synced?"
    log_info "Continue as requested"
  fi
}

system_upload_translations() {
  log_info "Uploading NGCP translations as requested..."
  if ! ngcp-prepare-translations ; then
    log_die "Failed to prepare ngcp-panel translations. Exiting."
    exit 1
  fi
}

system_allow_ssh_root() {
  echo -e "### Added by ngcp-installer\nPermitRootLogin yes" >> /etc/ssh/sshd_config
}

system_disable_services() {
  mkdir -p /etc/systemd/system-preset/
  local preset_file='/etc/systemd/system-preset/01-installer.preset'
  {
    # XXX: The following are not truly required by the installer, but due
    # to how they are currently declared, it causes problems on first reboot,
    # so we enable them for now here even if this is not strictly correct.
    echo "enable acpid.path"
    echo "enable acpid.socket"
    echo "enable ngcp-io-scheduler.service"
    # Minimum required services by the installer, during installation and
    # after the first reboot.
    echo "enable getty@.service"
    echo "enable networking.service"
    echo "enable procps.service"
    echo "enable remote-fs.target"
    echo "enable rsyslog.service"
    echo "enable ssh.service"
    echo "enable syslog.socket"
    echo "enable systemd-timesyncd.service"
    echo "disable *"
  } >> "${preset_file}"

  # These services do not have systemd unit file at this stage so they
  # are generated from /etc/init.d and can not be disabled via preset
  update-rc.d -f exim4 remove
  update-rc.d -f monit remove
  update-rc.d -f mysql remove
  update-rc.d -f openipmi remove
  update-rc.d -f prosody remove
  update-rc.d -f slapd remove

  systemctl preset-all
}

system_restart_network() {
  if [[ "${ORIGIN_INSTALL_DEV}" =~ ^(vlan|bond) ]]; then
    log_info "Deleting interface ${ORIGIN_INSTALL_DEV}"
    ip link set "${ORIGIN_INSTALL_DEV}" down
    ip link delete "${ORIGIN_INSTALL_DEV}"
  else
    ip addr flush dev "${ORIGIN_INSTALL_DEV}"
  fi
  log_info "Restarting network"
  systemctl restart networking
}

system_restart_ssh() {
  if "${DOCKER_BUILD}"; then
    log_info "It is Docker build, no need to restart ssh"
    return 0
  fi
  log_info "Restarting ssh daemon"
  systemctl restart ssh
}

system_save_configured_time() {
  echo "System configured. NGCP version ${VERSION} on $(date +"%Y-%m-%d %H:%M:%S")" >> /etc/sipwise_ngcp_version
  etckeeper commit "Changes for 'System configured' message"
}

system_create_swapfile() {
  local swapfile="/ngcp-data/swapfile"
  local swapfile_dir="$(dirname ${swapfile})"
  local swapsize_mb="${SWAPFILE_SIZE_MB}"

  swapsize_mb="${swapsize_mb%M*}"
  swapsize_mb="${swapsize_mb%m*}"

  if [[ "${swapsize_mb}" -eq 0 ]]; then
    log_info "Swapfile creation disabled"
    return
  fi

  if [[ -f "${swapfile}" ]]; then
    log_warn "Swapfile already exists: ${swapfile}"
    swapoff --all
    rm -f "${swapfile}"
  fi

  if [[ ! -d "${swapfile_dir}" ]]; then
    log_die "Dir for swapfile does not exist, or not a dir: ${swapfile_dir}"
    exit 1
  fi

  log_info "Creating swapfile at ${swapfile} with size ${swapsize_mb}MB ..."
  dd if=/dev/zero of="${swapfile}" bs=1M count="${swapsize_mb}" status=none
  chmod 600 "${swapfile}"
  mkswap "${swapfile}"
  if ! grep -q "${swapfile} none swap" /etc/fstab &>/dev/null ; then
    echo "${swapfile} none swap defaults 0 0" >> /etc/fstab
  fi

  log_info "Enabling swap:"
  swapon --all
  swapon --show

  echo "Reloading systemd manager configuration to inform about swap device"
  systemctl daemon-reload
}

system_run_ngcp_eaddress() {
  if ! "${CE}"; then
    log_info "Service ngcp-eaddress exists only on CE, skipping"
    return 0
  fi
  if "${DOCKER_BUILD}"; then
    log_info "It is Docker build, no need to restart ngcp-eaddress"
    return 0
  fi
  log_info "Restarting ngcp-eaddress"
  systemctl restart ngcp-eaddress
}

system_stop_all_services() {
  if [[ -f "/etc/ngcp-service/nsservices.yml" ]]; then
    ngcp-service --action=stop --group=ngcp-upgrade-essential || true
  fi
}

system_set_deploy_status() {
  [[ -n "$1" ]] || return 1
  echo "$*" > /tmp/status_server/status
}

system_enable_deploy_status_server() {
  local status_dir='/tmp/status_server'

  mkdir -p "${status_dir}"

  PY_MODULE='SimpleHTTPServer'
  if [[ $(python -V) =~ ^Python.*3 ]]; then
    PY_MODULE='http.server'
  fi
  export PY_MODULE
  # get rid of already running process
  local pid
  pid=$(pgrep -f "python.*${PY_MODULE}") || true
  [[ -n "${pid}" ]] && kill "${pid}"

  pushd "${status_dir}" &>/dev/null
  python -m "${PY_MODULE}" 4242 >/tmp/status_server.log 2>&1 &
  popd &>/dev/null
}

system_status_wait() {
  if [[ "${STATUS_WAIT_SECONDS}" != 0 ]]; then
    # wait for a specific number of seconds so a remote host has
    # a chance to check for deploy status "finished"
    echo "Sleeping for ${STATUS_WAIT_SECONDS} seconds (as requested via option 'STATUS_WAIT_SECONDS')"
    sleep "${STATUS_WAIT_SECONDS}"
  fi
}

system_wait_exit() {
  local e_code="${?}"
  if [[ "${e_code}" -eq 0 ]]; then
    system_set_deploy_status "finished"
  else
    system_set_deploy_status "error"
  fi
  trap '' 1 2 3 6 15 ERR EXIT
  system_status_wait
  local pid
  pid=$(pgrep -f "python.*${PY_MODULE}") || true
  [[ -n "${pid}" ]] && kill "${pid}"
  exit "${e_code}"
}

system_delete_bootstrap_key() {
  local sipwise_key='/etc/apt/trusted.gpg.d/sipwise-keyring-bootstrap.gpg'
  local status
  status="$(dpkg-query -f "\${db:Status-Status} \${db:Status-Eflag}" -W ngcp-archive-keyring 2>/dev/null)"
  if [[ -f "${sipwise_key}" ]]; then
    if [[ "${status}" != 'installed ok' ]]; then
      log_die "There is '${sipwise_key}' file but ngcp-archive-keyring package is not installed"
      log_die "This case should not happen normally, exiting..."
      exit 1
    fi
    rm -f "${sipwise_key}"
  else
    if [[ "${status}" != 'installed ok' ]]; then
      log_die "There is neither '${sipwise_key}' file nor ngcp-archive-keyring package"
      log_die "This case should not happen normally, exiting..."
      exit 1
    fi
  fi
}

system_usage() {
  cat - <<__USAGE__
$(basename "$0") used to perform the initial configuration of Sipwise C5 system.

Usage: $(basename "$0") [options]

Important note: all the settings will be taken from /etc/ngcp-installer/config_deploy.inc,
                edit it for fine tuning if necessary.

Options:

    -j|--join                - Join the cluster, all yml configs are retrieved from management node.
                               This option should not be used along with --use-existing-yml.

    --docker                 - Used to configure ngcp system within docker container.
                               Some actions, like network configuration and systemd calls are
                               skipped.

    --use-existing-yml=value - Skip the creation of some yml files and use the existing ones
                               from /etc/ngcp-config.
                               This option is used only for installation of the first management
                               node and should not be used along with --join.

                               Multiple values are possible, should be separated with comma:
        config               - use existing config.yml
        constants            - use existing constants.yml
        network              - use existing network.yml

    -h|--help                - Show this message.

__USAGE__
}

system_remove_ngcp_config_from_git() {
  pushd /etc &>/dev/null
  echo "#ignore ngcp configuration files" >> .gitignore
  echo "ngcp-config" >> .gitignore
  git add .gitignore
  git rm --cached -r ngcp-config
  git commit -m 'Ignore ngcp-config files'
  popd &>/dev/null
}
